"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Text = exports.stringToArray = void 0;
const Util_1 = require("../Util");
const Factory_1 = require("../Factory");
const Shape_1 = require("../Shape");
const linkify = require("linkifyjs");
const Validators_1 = require("../Validators");
const Global_1 = require("../Global");
const DEFAULT_LINK_COLOR = "#419bf9";
function stringToArray(string) {
    return Array.from(string);
}
exports.stringToArray = stringToArray;
var AUTO = 'auto', CENTER = 'center', JUSTIFY = 'justify', CHANGE_KONVA = 'Change.konva', CONTEXT_2D = '2d', DASH = '-', LEFT = 'left', TEXT = 'text', TEXT_UPPER = 'Text', TOP = 'top', BOTTOM = 'bottom', MIDDLE = 'middle', NORMAL = 'normal', PX_SPACE = 'px ', SPACE = ' ', RIGHT = 'right', WORD = 'word', CHAR = 'char', NONE = 'none', ELLIPSIS = '…', ATTR_CHANGE_LIST = [
    'fontFamily',
    'fontSize',
    'fontStyle',
    'fontVariant',
    'padding',
    'align',
    'verticalAlign',
    'lineHeight',
    'text',
    'width',
    'height',
    'wrap',
    'ellipsis',
    'letterSpacing',
    'linkFill',
], attrChangeListLen = ATTR_CHANGE_LIST.length;
function normalizeFontFamily(fontFamily) {
    return fontFamily
        .split(',')
        .map((family) => {
        family = family.trim();
        const hasSpace = family.indexOf(' ') >= 0;
        const hasQuotes = family.indexOf('"') >= 0 || family.indexOf("'") >= 0;
        if (hasSpace && !hasQuotes) {
            family = `"${family}"`;
        }
        return family;
    })
        .join(', ');
}
var dummyContext;
function getDummyContext() {
    if (dummyContext) {
        return dummyContext;
    }
    dummyContext = Util_1.Util.createCanvasElement().getContext(CONTEXT_2D);
    return dummyContext;
}
function _fillFunc(context) {
    context.fillStyle = this.fill();
    const linkLine = this.textArr.find(textLine => (this._partialText === textLine.text &&
        textLine.isLink));
    if (linkLine && !linkLine.hrefPosition) {
        context.fillStyle = this.linkFill() || DEFAULT_LINK_COLOR;
        context.fillText(this._partialText, this._partialTextX, this._partialTextY);
        return;
    }
    if (linkLine && linkLine.hrefPosition) {
        const partialLinkText = this._partialText.slice(linkLine.hrefPosition[0], linkLine.hrefPosition[1]);
        const initialPartialText = this._partialText.slice(0, linkLine.hrefPosition[0]);
        const lastPartialText = this._partialText.slice(linkLine.hrefPosition[1], this._partialText.length);
        if (initialPartialText.length) {
            const partialTextX = this._calculateStrokeInitialPos(linkLine.width);
            context.fillStyle = this.fill();
            context.fillText(initialPartialText, partialTextX, this._partialTextY);
        }
        if (lastPartialText.length) {
            context.fillStyle = this.fill();
            context.fillText(lastPartialText, linkLine.xEnd, this._partialTextY);
        }
        context.fillStyle = this.linkFill() || DEFAULT_LINK_COLOR;
        context.fillText(partialLinkText, linkLine.xStart, this._partialTextY);
        return;
    }
    context.fillText(this._partialText, this._partialTextX, this._partialTextY);
}
function _strokeFunc(context) {
    context.setAttr('miterLimit', 2);
    context.strokeText(this._partialText, this._partialTextX, this._partialTextY);
}
function checkDefaultFill(config) {
    config = config || {};
    if (!config.fillLinearGradientColorStops &&
        !config.fillRadialGradientColorStops &&
        !config.fillPatternImage) {
        config.fill = config.fill || 'black';
    }
    return config;
}
class Text extends Shape_1.Shape {
    constructor(config) {
        var _a;
        super(checkDefaultFill(config));
        this._partialTextX = 0;
        this._partialTextY = 0;
        this._findMethod = (config === null || config === void 0 ? void 0 : config.displayLinks) && ((_a = config === null || config === void 0 ? void 0 : config.findMethod) !== null && _a !== void 0 ? _a : linkify.find);
        this._linkLines = [];
        for (var n = 0; n < attrChangeListLen; n++) {
            this.on(ATTR_CHANGE_LIST[n] + CHANGE_KONVA, this._setTextData);
        }
        this._setTextData();
    }
    _calculateStrokeInitialPos(linkWidth) {
        var _a, _b;
        if (this.align() === LEFT) {
            return 0;
        }
        const plus = (this.getWidth() - linkWidth + (((_a = this.offsetX()) !== null && _a !== void 0 ? _a : 0) / 2)) / 2;
        const minus = (this.getWidth() - linkWidth - (((_b = this.offsetX()) !== null && _b !== void 0 ? _b : 0) / 2)) / 2;
        return (plus + minus) / 2;
    }
    _sceneFunc(context) {
        var textArr = this.textArr, textArrLen = textArr.length;
        if (!this.text()) {
            return;
        }
        var padding = this.padding(), fontSize = this.fontSize(), lineHeightPx = this.lineHeight() * fontSize, verticalAlign = this.verticalAlign(), alignY = 0, align = this.align(), totalWidth = this.getWidth(), letterSpacing = this.letterSpacing(), fill = this.fill(), textDecoration = this.textDecoration(), shouldUnderline = textDecoration.indexOf('underline') !== -1, shouldLineThrough = textDecoration.indexOf('line-through') !== -1, n;
        var translateY = 0;
        var translateY = lineHeightPx / 2;
        var lineTranslateX = 0;
        var lineTranslateY = 0;
        context.setAttr('font', this._getContextFont());
        context.setAttr('textBaseline', MIDDLE);
        context.setAttr('textAlign', LEFT);
        if (verticalAlign === MIDDLE) {
            alignY = (this.getHeight() - textArrLen * lineHeightPx - padding * 2) / 2;
        }
        else if (verticalAlign === BOTTOM) {
            alignY = this.getHeight() - textArrLen * lineHeightPx - padding * 2;
        }
        context.translate(padding, alignY + padding);
        for (n = 0; n < textArrLen; n++) {
            var lineTranslateX = 0;
            var lineTranslateY = 0;
            var obj = textArr[n], text = obj.text, width = obj.width, lastLine = obj.lastInParagraph, isLink = obj.isLink, xStart = obj.xStart, xEnd = obj.xEnd, spacesNumber, oneWord, lineWidth;
            context.save();
            if (align === RIGHT) {
                lineTranslateX += totalWidth - width - padding * 2;
            }
            else if (align === CENTER) {
                lineTranslateX += (totalWidth - width - padding * 2) / 2;
            }
            if (shouldUnderline) {
                context.save();
                context.beginPath();
                context.moveTo(lineTranslateX, translateY + lineTranslateY + Math.round(fontSize / 2));
                spacesNumber = text.split(' ').length - 1;
                oneWord = spacesNumber === 0;
                lineWidth =
                    align === JUSTIFY && !lastLine ? totalWidth - padding * 2 : width;
                context.lineTo(lineTranslateX + Math.round(lineWidth), translateY + lineTranslateY + Math.round(fontSize / 2));
                context.lineWidth = fontSize / 15;
                const gradient = this._getLinearGradient();
                context.strokeStyle = gradient || fill;
                context.stroke();
                context.restore();
            }
            if (isLink) {
                context.save();
                context.beginPath();
                context.moveTo(xStart, translateY + lineTranslateY + Math.round(fontSize / 2));
                lineWidth = align === JUSTIFY && !lastLine ? totalWidth - padding * 2 : width;
                context.lineTo(xEnd, translateY + lineTranslateY + Math.round(fontSize / 2));
                context.lineWidth = fontSize / 15;
                context.fillStyle = this.linkFill() || DEFAULT_LINK_COLOR;
                context.strokeStyle = this.linkFill() || DEFAULT_LINK_COLOR;
                context.stroke();
                context.fill();
                context.restore();
            }
            if (shouldLineThrough) {
                context.save();
                context.beginPath();
                context.moveTo(lineTranslateX, translateY + lineTranslateY);
                spacesNumber = text.split(' ').length - 1;
                oneWord = spacesNumber === 0;
                lineWidth =
                    align === JUSTIFY && lastLine && !oneWord
                        ? totalWidth - padding * 2
                        : width;
                context.lineTo(lineTranslateX + Math.round(lineWidth), translateY + lineTranslateY);
                context.lineWidth = fontSize / 15;
                const gradient = this._getLinearGradient();
                context.strokeStyle = gradient || fill;
                context.stroke();
                context.restore();
            }
            if (letterSpacing !== 0 || align === JUSTIFY) {
                spacesNumber = text.split(' ').length - 1;
                var array = stringToArray(text);
                for (var li = 0; li < array.length; li++) {
                    var letter = array[li];
                    if (letter === ' ' && !lastLine && align === JUSTIFY) {
                        lineTranslateX += (totalWidth - padding * 2 - width) / spacesNumber;
                    }
                    this._partialTextX = lineTranslateX;
                    this._partialTextY = translateY + lineTranslateY;
                    this._partialText = letter;
                    context.fillStrokeShape(this);
                    lineTranslateX += this.measureSize(letter).width + letterSpacing;
                }
            }
            else {
                this._partialTextX = lineTranslateX;
                this._partialTextY = translateY + lineTranslateY;
                this._partialText = text;
                context.fillStrokeShape(this);
            }
            context.restore();
            if (textArrLen > 1) {
                translateY += lineHeightPx;
            }
        }
    }
    _hitFunc(context) {
        var width = this.getWidth(), height = this.getHeight();
        context.beginPath();
        context.rect(0, 0, width, height);
        context.closePath();
        context.fillStrokeShape(this);
    }
    setText(text) {
        var str = Util_1.Util._isString(text)
            ? text
            : text === null || text === undefined
                ? ''
                : text + '';
        this._setAttr(TEXT, str);
        return this;
    }
    getWidth() {
        var isAuto = this.attrs.width === AUTO || this.attrs.width === undefined;
        return isAuto ? this.getTextWidth() + this.padding() * 2 : this.attrs.width;
    }
    getHeight() {
        var isAuto = this.attrs.height === AUTO || this.attrs.height === undefined;
        return isAuto
            ? this.fontSize() * this.textArr.length * this.lineHeight() +
                this.padding() * 2
            : this.attrs.height;
    }
    getTextWidth() {
        return this.textWidth;
    }
    getTextHeight() {
        Util_1.Util.warn('text.getTextHeight() method is deprecated. Use text.height() - for full height and text.fontSize() - for one line height.');
        return this.textHeight;
    }
    measureSize(text) {
        var _context = getDummyContext(), fontSize = this.fontSize(), metrics;
        _context.save();
        _context.font = this._getContextFont();
        metrics = _context.measureText(text);
        _context.restore();
        return {
            width: metrics.width,
            height: fontSize,
        };
    }
    _getContextFont() {
        return (this.fontStyle() +
            SPACE +
            this.fontVariant() +
            SPACE +
            (this.fontSize() + PX_SPACE) +
            normalizeFontFamily(this.fontFamily()));
    }
    _addTextLine(line, autoNextLine = false) {
        var _a, _b, _c;
        const align = this.align();
        if (align === JUSTIFY) {
            line = line.trim();
        }
        var width = this._getTextWidth(line);
        const maybeLinked = (_b = (_a = this._findMethod) === null || _a === void 0 ? void 0 : _a.call(this, line)) !== null && _b !== void 0 ? _b : [];
        const detectLink = (maybeLinked !== null && maybeLinked !== void 0 ? maybeLinked : []).some((result) => result.isLink);
        const previousLine = this.textArr[this.textArr.length - 1];
        const isLink = !autoNextLine
            ? detectLink
            : detectLink || previousLine && (previousLine.isLink && (!previousLine.hrefPosition ||
                previousLine.text.length === previousLine.hrefPosition[1]));
        let hrefPosition = null;
        if (isLink && !autoNextLine && line !== ((_c = maybeLinked[0]) === null || _c === void 0 ? void 0 : _c.value)) {
            hrefPosition = [maybeLinked[0].start, maybeLinked[0].end];
        }
        if (isLink && autoNextLine && !maybeLinked.length && line.split(" ").length > 1) {
            hrefPosition = [0, line.split(" ")[0].length];
        }
        const { xStart, xEnd } = this._getTextPosition(width, hrefPosition, line);
        return this.textArr.push({
            text: line,
            width: width,
            isLink,
            autoNextLine: (!!autoNextLine && !detectLink),
            hrefPosition,
            xStart,
            xEnd,
            lastInParagraph: false
        });
    }
    _getTextPosition(width, hrefPosition, text) {
        var _a, _b, _c, _d;
        const xInitial = this._calculateStrokeInitialPos(width);
        if (!hrefPosition) {
            return {
                xStart: xInitial,
                xEnd: xInitial + width,
            };
        }
        const startPadding = (_b = (_a = getDummyContext().measureText(text.slice(0, hrefPosition[0]))) === null || _a === void 0 ? void 0 : _a.width) !== null && _b !== void 0 ? _b : 0;
        const endPadding = (_d = (_c = getDummyContext().measureText(text.slice(hrefPosition[0], hrefPosition[1]))) === null || _c === void 0 ? void 0 : _c.width) !== null && _d !== void 0 ? _d : 0;
        return {
            xStart: xInitial + startPadding,
            xEnd: xInitial + startPadding + endPadding
        };
    }
    _addLinePositions() {
        const textArr = this.textArr;
        const lineHeightPx = this.lineHeight() * this.fontSize();
        const yInitial = (this.getHeight() - (textArr.length * lineHeightPx)) / 2;
        let linkOrder = -1;
        this.textArr = this.textArr.map((textLine, idx) => {
            var _a;
            if (!textLine.isLink) {
                return textLine;
            }
            if (textLine.isLink && !textLine.autoNextLine) {
                linkOrder += 1;
            }
            return Object.assign(Object.assign({}, textLine), { yStart: yInitial + (lineHeightPx * idx), yEnd: yInitial + (lineHeightPx * (idx + 1)), fullLink: (_a = this._linkLines[linkOrder]) === null || _a === void 0 ? void 0 : _a.href });
        });
    }
    _getTextWidth(text) {
        var letterSpacing = this.letterSpacing();
        var length = text.length;
        return (getDummyContext().measureText(text).width +
            (length ? letterSpacing * (length - 1) : 0));
    }
    _setTextData() {
        var _a, _b, _c, _d, _e, _f, _g;
        var lines = this.text().split('\n'), fontSize = +this.fontSize(), textWidth = 0, lineHeightPx = this.lineHeight() * fontSize, width = this.attrs.width, height = this.attrs.height, fixedWidth = width !== AUTO && width !== undefined, fixedHeight = height !== AUTO && height !== undefined, padding = this.padding(), maxWidth = width - padding * 2, maxHeightPx = height - padding * 2, currentHeightPx = 0, wrap = this.wrap(), shouldWrap = wrap !== NONE, wrapAtWord = wrap !== CHAR && shouldWrap, shouldAddEllipsis = this.ellipsis();
        this.textArr = [];
        this._linkLines = [];
        getDummyContext().font = this._getContextFont();
        var additionalWidth = shouldAddEllipsis ? this._getTextWidth(ELLIPSIS) : 0;
        for (var i = 0, max = lines.length; i < max; ++i) {
            var line = lines[i];
            const linkLine = ((_b = (_a = this._findMethod) === null || _a === void 0 ? void 0 : _a.call(this, line)) !== null && _b !== void 0 ? _b : []).filter((result) => result.isLink);
            if (linkLine.length) {
                this._linkLines.push(...linkLine);
            }
            var lineWidth = this._getTextWidth(line);
            if (fixedWidth && lineWidth > maxWidth) {
                while (line.length > 0) {
                    var low = 0, high = line.length, match = '', matchWidth = 0;
                    while (low < high) {
                        var mid = (low + high) >>> 1, substr = line.slice(0, mid + 1), substrWidth = this._getTextWidth(substr) + additionalWidth;
                        if (substrWidth <= maxWidth) {
                            low = mid + 1;
                            match = substr;
                            matchWidth = substrWidth;
                        }
                        else {
                            high = mid;
                        }
                    }
                    if (match) {
                        if (wrapAtWord) {
                            var wrapIndex;
                            var nextChar = line[match.length];
                            var nextIsSpaceOrDash = nextChar === SPACE || nextChar === DASH;
                            if (nextIsSpaceOrDash && matchWidth <= maxWidth) {
                                wrapIndex = match.length;
                            }
                            else {
                                wrapIndex =
                                    Math.max(match.lastIndexOf(SPACE), match.lastIndexOf(DASH)) +
                                        1;
                            }
                            if (wrapIndex > 0) {
                                low = wrapIndex;
                                match = match.slice(0, low);
                                matchWidth = this._getTextWidth(match);
                            }
                        }
                        match = match.trimRight();
                        const linesMaybeLinkPart = (_e = ((_d = (_c = this._findMethod) === null || _c === void 0 ? void 0 : _c.call(this, lines[i])) !== null && _d !== void 0 ? _d : [])[0]) === null || _e === void 0 ? void 0 : _e.value;
                        const maybeFoundLinkPart = (_g = (_f = this._linkLines) === null || _f === void 0 ? void 0 : _f.find(someLine => someLine.value === linesMaybeLinkPart)) === null || _g === void 0 ? void 0 : _g.value;
                        const isAutoNextLine = lines[i].includes(match) && maybeFoundLinkPart && !match.includes(maybeFoundLinkPart);
                        this._addTextLine(match, isAutoNextLine);
                        textWidth = Math.max(textWidth, matchWidth);
                        currentHeightPx += lineHeightPx;
                        var shouldHandleEllipsis = this._shouldHandleEllipsis(currentHeightPx);
                        if (shouldHandleEllipsis) {
                            this._tryToAddEllipsisToLastLine();
                            break;
                        }
                        let trimmedSpace = line.slice(low - 1, low) === " ";
                        line = line.slice(low);
                        const initialLength = line.length;
                        line = line.trimLeft();
                        trimmedSpace = trimmedSpace || (line.length !== initialLength);
                        if (line.length > 0) {
                            lineWidth = this._getTextWidth(line);
                            if (lineWidth <= maxWidth) {
                                this._addTextLine(line, !trimmedSpace);
                                currentHeightPx += lineHeightPx;
                                textWidth = Math.max(textWidth, lineWidth);
                                break;
                            }
                        }
                    }
                    else {
                        break;
                    }
                }
            }
            else {
                this._addTextLine(line);
                currentHeightPx += lineHeightPx;
                textWidth = Math.max(textWidth, lineWidth);
                if (this._shouldHandleEllipsis(currentHeightPx) && i < max - 1) {
                    this._tryToAddEllipsisToLastLine();
                }
            }
            if (this.textArr[this.textArr.length - 1]) {
                this.textArr[this.textArr.length - 1].lastInParagraph = true;
            }
            if (fixedHeight && currentHeightPx + lineHeightPx > maxHeightPx) {
                break;
            }
        }
        this.textHeight = fontSize;
        this.textWidth = textWidth;
        this._addLinePositions();
    }
    _shouldHandleEllipsis(currentHeightPx) {
        var fontSize = +this.fontSize(), lineHeightPx = this.lineHeight() * fontSize, height = this.attrs.height, fixedHeight = height !== AUTO && height !== undefined, padding = this.padding(), maxHeightPx = height - padding * 2, wrap = this.wrap(), shouldWrap = wrap !== NONE;
        return (!shouldWrap ||
            (fixedHeight && currentHeightPx + lineHeightPx > maxHeightPx));
    }
    _tryToAddEllipsisToLastLine() {
        var width = this.attrs.width, fixedWidth = width !== AUTO && width !== undefined, padding = this.padding(), maxWidth = width - padding * 2, shouldAddEllipsis = this.ellipsis();
        var lastLine = this.textArr[this.textArr.length - 1];
        if (!lastLine || !shouldAddEllipsis) {
            return;
        }
        if (fixedWidth) {
            var haveSpace = this._getTextWidth(lastLine.text + ELLIPSIS) < maxWidth;
            if (!haveSpace) {
                lastLine.text = lastLine.text.slice(0, lastLine.text.length - 3);
            }
        }
        this.textArr.splice(this.textArr.length - 1, 1);
        this._addTextLine(lastLine.text + ELLIPSIS, lastLine.autoNextLine);
    }
    getStrokeScaleEnabled() {
        return true;
    }
}
exports.Text = Text;
Text.prototype._fillFunc = _fillFunc;
Text.prototype._strokeFunc = _strokeFunc;
Text.prototype.className = TEXT_UPPER;
Text.prototype._attrsAffectingSize = [
    'text',
    'fontSize',
    'padding',
    'wrap',
    'lineHeight',
    'letterSpacing',
];
(0, Global_1._registerNode)(Text);
Factory_1.Factory.overWriteSetter(Text, 'width', (0, Validators_1.getNumberOrAutoValidator)());
Factory_1.Factory.overWriteSetter(Text, 'height', (0, Validators_1.getNumberOrAutoValidator)());
Factory_1.Factory.addGetterSetter(Text, 'fontFamily', 'Arial');
Factory_1.Factory.addGetterSetter(Text, 'fontSize', 12, (0, Validators_1.getNumberValidator)());
Factory_1.Factory.addGetterSetter(Text, 'fontStyle', NORMAL);
Factory_1.Factory.addGetterSetter(Text, 'fontVariant', NORMAL);
Factory_1.Factory.addGetterSetter(Text, 'padding', 0, (0, Validators_1.getNumberValidator)());
Factory_1.Factory.addGetterSetter(Text, 'align', LEFT);
Factory_1.Factory.addGetterSetter(Text, 'linkFill', undefined, (0, Validators_1.getStringOrGradientValidator)());
Factory_1.Factory.addGetterSetter(Text, 'verticalAlign', TOP);
Factory_1.Factory.addGetterSetter(Text, 'lineHeight', 1, (0, Validators_1.getNumberValidator)());
Factory_1.Factory.addGetterSetter(Text, 'wrap', WORD);
Factory_1.Factory.addGetterSetter(Text, 'ellipsis', false, (0, Validators_1.getBooleanValidator)());
Factory_1.Factory.addGetterSetter(Text, 'letterSpacing', 0, (0, Validators_1.getNumberValidator)());
Factory_1.Factory.addGetterSetter(Text, 'text', '', (0, Validators_1.getStringValidator)());
Factory_1.Factory.addGetterSetter(Text, 'textDecoration', '');
